---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

-- Create a new SpellBook class
---@class Bastion.SpellBook
---@field spells table<number, Bastion.Spell>
---@field auras { [spellId]: { [spellId]: { spell: Bastion.Spell, source?: Bastion.Unit, target?: Bastion.Unit, lastApplied: number } } }
local SpellBook = {}
SpellBook.__index = SpellBook

-- Constructor
---@param global? boolean
---@return Bastion.SpellBook
function SpellBook:New(global)
    ---@class Bastion.SpellBook
    local self = setmetatable({}, SpellBook)
    self.spells = {}
    self.global = global or false
    --self:RegisterEvents()
    return self
end

function SpellBook:RegisterEvents()
    Bastion.Globals.EventManager:RegisterWoWEvent(
        { "UNIT_SPELLCAST_SENT", "UNIT_SPELLCAST_SUCCEEDED", "UNIT_SPELLCAST_FAILED", "UNIT_SPELLCAST_FAILED_QUIET",
            "UNIT_SPELLCAST_INTERRUPTED", "UNIT_SPELLCAST_START" },
        function(...)
            self:UnitSpellCast(...)
        end, true)
end

---@param event "UNIT_SPELLCAST_SENT" | "UNIT_SPELLCAST_SUCCEEDED" | "UNIT_SPELLCAST_FAILED" | "UNIT_SPELLCAST_FAILED_QUIET" | "UNIT_SPELLCAST_INTERRUPTED"
function SpellBook:UnitSpellCast(event, ...)
    local casterUnitId, targetUnitId, castGUID, spellId
    local currentTime = GetTime()
    if event == "UNIT_SPELLCAST_SENT" then
        ---@type UnitId, string, string, number
        casterUnitId, targetUnitId, castGUID, spellId = ...
    else
        ---@type UnitId, string, number
        casterUnitId, castGUID, spellId = ...
    end

    local spell = self:GetIfRegistered(spellId) or self.global and self:GetSpell(spellId)
    if spell then
        if event == "UNIT_SPELLCAST_SENT" then
            spell:SetLastCastAttempt(GetTime(), castGUID)
            -- print(event, self.global, ...)
            local casterUnit = Bastion.Globals.UnitManager:GetObject(casterUnitId)
            if casterUnit and casterUnit:IsValid() and casterUnit.castGUID ~= castGUID then
                casterUnit:SetLastSpell(spell, castGUID)
            end
        elseif event == "UNIT_SPELLCAST_SUCCEEDED" then
            spell:SetLastCast(GetTime(), true, castGUID)
            -- print(event, self.global, ...)
            local casterUnit = Bastion.Globals.UnitManager:GetObject(casterUnitId)
            if casterUnit and casterUnit:IsValid() then
                casterUnit:SetSpell(false, castGUID)
                casterUnit:SetLastCastInterrupted(false)
            end
        elseif event == "UNIT_SPELLCAST_FAILED" or event == "UNIT_SPELLCAST_FAILED_QUIET" then
            if spell.castGUID == castGUID then
                spell:SetLastCast(GetTime(), false, castGUID)
            end
            -- print(event, self.global, ...)
            local casterUnit = Bastion.Globals.UnitManager:GetObject(casterUnitId)
            if casterUnit and casterUnit:IsValid() then
                casterUnit:SetSpell(false)
            end
        elseif event == "UNIT_SPELLCAST_INTERRUPTED" then
            -- print(event, self.global, ...)
            local casterUnit = Bastion.Globals.UnitManager:GetObject(casterUnitId)
            if casterUnit and casterUnit:IsValid() then
                casterUnit:SetSpell(false, castGUID)
                casterUnit:SetLastCastInterrupted(true)
            end
        end
    end
end

-- Get a spell from the spellbook
---@param id integer|string
---@return Bastion.Spell
function SpellBook:GetSpell(id)
    local override = FindSpellOverrideByID(id)
    if self.spells[id] == nil then
        self.spells[id] = Bastion.Spell:New(id)
    end

    if override and override ~= id then
        if self.spells[override] == nil then
            self.spells[override] = Bastion.Spell:New(override)
        end
    end

    return self.spells[id]
end

---@param ... integer
---@return Bastion.Spell, ... Spell
function SpellBook:GetSpells(...)
    local spells = {}
    for _, id in ipairs({ ... }) do
        table.insert(spells, self:GetSpell(id))
    end

    return unpack(spells)
end

---@param aura Bastion.Spell
---@return Bastion.Spell[]
function SpellBook:GetAuraSpells(aura)
    local spells = {}
    for _, spell in pairs(self.spells) do
        if spell:CheckAuraStatus(aura) then
            table.insert(spells, spell)
        end
    end

    return spells
end

---@param ... integer
---@return Bastion.List
function SpellBook:GetList(...)
    local spells = {}
    for _, id in ipairs({ ... }) do
        table.insert(spells, self:GetSpell(id))
    end

    return Bastion.List:New(spells)
end

---@param name string
---@return Bastion.Spell
function SpellBook:GetSpellByName(name)
    local spellInfo = C_Spell.GetSpellInfo(name)
    return self:GetSpell(spellInfo and spellInfo.spellID or 0)
end

---@param id integer
function SpellBook:GetIfRegistered(id)
    if self.spells[id] then
        return self.spells[id]
    end

    if self.spells[FindSpellOverrideByID(id)] then
        return self.spells[FindSpellOverrideByID(id)]
    end

    return false
end

Bastion.SpellBook = SpellBook
--Bastion.Globals.SpellBook = Bastion.SpellBook:New(true)
